/*****************************************************************************\
*              efs - General purpose Embedded Filesystem library              *
*          --------------------- -----------------------------------          *
*                                                                             *
* Filename : hcs12cw_spi.c                                                    *
* Description : This file contains the functions needed to use efs for        *
*               accessing files on an SD-card connected to an Freescale       *
*               HCS12 via SPI.                                                *
*                                                                             *
* This program is free software; you can redistribute it and/or               *
* modify it under the terms of the GNU General Public License                 *
* as published by the Free Software Foundation; version 2                     *
* of the License.                                                             *
                                                                              *
* This program is distributed in the hope that it will be useful,             *
* but WITHOUT ANY WARRANTY; without even the implied warranty of              *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               *
* GNU General Public License for more details.                                *
*                                                                             *
* As a special exception, if other files instantiate templates or             *
* use macros or inline functions from this file, or you compile this          *
* file and link it with other works to produce a work based on this file,     *
* this file does not by itself cause the resulting work to be covered         *
* by the GNU General Public License. However the source code for this         *
* file must still be made available in accordance with section (3) of         *
* the GNU General Public License.                                             *
*                                                                             *
* This exception does not invalidate any other reasons why a work based       *
* on this file might be covered by the GNU General Public License.            *
*                                                                             *
*                                                    (c)2006 Lennart Yseboodt *
*                                                    (c)2006 Michael De Nil   *
*                                                    (c)2007 Konstantin Les   *
\*****************************************************************************/


#include <hidef.h>      /* common defines and macros */
#include <mc9s12dg128.h>     /* derivative information */
#pragma LINK_INFO DERIVATIVE "mc9s12dg128b"

#if 0
#include <hidef.h>      /* common defines and macros */
#include <mc9s12xdt256.h>     /* derivative information */
#pragma LINK_INFO DERIVATIVE "mc9s12xdt256"
#endif

#include "interfaces/hcs12cw_spi.h"
#include "interfaces/sd.h"


/*================================================================================================*/
#define QNT_SPI (2)

#define SPI_DEVNUM_UNDEFINED (255)

#define SS_MODE_NA (0)
#define SS_MODE_HW (1)
#define SS_MODE_SW (2)


#define NUMBER_ARRAY_ELEMENTS(a) (sizeof(a) / sizeof((a)[0]))

#define GET_SPI0(b) do {while (!SPI0SR_SPIF) ; (b) = SPI0DR;} while(0);
#define GET_SPI1(b) do {while (!SPI1SR_SPIF) ; (b) = SPI1DR;} while(0);
#define PUT_SPI0(b) do {while (!SPI0SR_SPTEF) ; __asm nop; __asm nop; SPI0DR = (b);} while(0);
#define PUT_SPI1(b) do {while (!SPI1SR_SPTEF) ; __asm nop; __asm nop; SPI1DR = (b);} while(0);
/*================================================================================================*/

/*================================================================================================*/
struct spi_config
    {
    unsigned char ss_mode;
    unsigned char *ss_port_addr;
    unsigned char ss_pin_mask;
    unsigned char transfer_mode;
    };
struct spidev_config
    {
    unsigned char spi_number;
    struct spi_config spi_mode;
    };
/*================================================================================================*/

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
#include "if_cfg.c"
#define QNT_SPI_DEV (NUMBER_ARRAY_ELEMENTS( spi_devices ))
/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

/*================================================================================================*/
static void f_initialize_SPI0(const struct spi_config *spi_mode, unsigned char br_val);
static euint8 f_send_SPI0(const struct spi_config *spi_mode, euint8 outgoing);
static void f_sendbuf_SPI0(const struct spi_config *spi_mode, euint16 bufsize, const euint8 *buf);
/*================================================================================================*/

/*================================================================================================*/
static unsigned char f_cur_device_number = SPI_DEVNUM_UNDEFINED;
static unsigned char f_device_baud[ QNT_SPI_DEV ];
static const struct
    {
    void (*init)(const struct spi_config *, euint8);
    euint8 (*send)(const struct spi_config *, euint8);
    void (*sendbuf)(const struct spi_config *, euint16, const euint8 *);
    } f_if_spi_handles[ QNT_SPI ] = 
    {
    {f_initialize_SPI0, f_send_SPI0, f_sendbuf_SPI0},
    {NULL, NULL, NULL}
    };
/*================================================================================================*/

/*================================================================================================*/
/*------------------------------------------------------------------------------------------------*/
static void f_initialize_SPI0
    (
    const struct spi_config *spi_mode, /* (in) - SPI mode */
    euint8 br_val                      /* (in) - value of SPI0BR register -> SPI baud rate */
    )
    {
    int spi_master_ok;
    
    if (if_take_SPI0() != 0)
        return;
    
    SPI0CR1_CPHA = 0;
    SPI0CR1_SPE = 0;
    SPI0BR = br_val;
    SPI0CR2 = (spi_mode->ss_mode == SS_MODE_HW) ? SPI0CR2_MODFEN_MASK : 0;
    do
        {
        SPI0CR1_CPHA = 0;
        
        if (spi_mode->ss_mode == SS_MODE_HW)
            {
            unsigned char spi_status;
            SPI0CR1 = SPI0CR1_SPE_MASK | SPI0CR1_MSTR_MASK | SPI0CR1_SSOE_MASK |
                ((spi_mode->transfer_mode & 0x03) << 2);
            spi_status = SPI0SR;
            spi_master_ok = ((spi_status & SPI0SR_MODF_MASK) == 0);
            }
        else
            {
            SPI0CR1 = SPI0CR1_SPE_MASK | SPI0CR1_MSTR_MASK |
                ((spi_mode->transfer_mode & 0x03) << 2);
            spi_master_ok = 1;
            }
        }
    while (!spi_master_ok);
    
    if_release_SPI0();
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
static euint8 f_send_SPI0
    (
    const struct spi_config *spi_mode,
    euint8 outgoing
    )
    {
    euint8 incoming = 0;
    
    if (if_take_SPI0() != 0)
        return 0;
    
    /* SS pin functionality */
    if (spi_mode->ss_mode == SS_MODE_SW)
        *(spi_mode->ss_port_addr) &= ~spi_mode->ss_pin_mask;
    
    PUT_SPI0( outgoing );
    GET_SPI0( incoming );
    
    /* SS pin functionality */
    if (spi_mode->ss_mode == SS_MODE_SW)
        *(spi_mode->ss_port_addr) |= spi_mode->ss_pin_mask;
    
    if_release_SPI0();
    
    return incoming;
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
static void f_sendbuf_SPI0
    (
    const struct spi_config *spi_mode,
    euint16 bufsize,
    const euint8 *buf
    )
    {
    unsigned char rx_byte;
    
    if (bufsize == 0)
        return;
    
    if (if_take_SPI0() != 0)
        return;
    
    
    /* SS pin functionality */
    if (spi_mode->ss_mode == SS_MODE_SW)
        *(spi_mode->ss_port_addr) &= ~spi_mode->ss_pin_mask;
    
    while (bufsize-- > 1)
        {
        PUT_SPI0( *buf );
        buf++;
        }
    if_enter_critical();
    PUT_SPI0( *buf );
    while (!SPI0SR_SPTEF)
        {
        continue;
        }
    rx_byte = SPI0SR;
    rx_byte = SPI0DR;
    if_exit_critical();
    
    GET_SPI0( rx_byte );

    /* SS pin functionality */
    if (spi_mode->ss_mode == SS_MODE_SW)
        *(spi_mode->ss_port_addr) |= spi_mode->ss_pin_mask;
    
    if_release_SPI0();
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
esint8 if_initInterface
    (
    hwInterface *file,
    eint8 *opts                        /* (in) - used SPI device number */
    )
    {
    euint32 sc;
    
    if (*opts >= QNT_SPI_DEV)
        return -1;
    
    file->numDevice = *opts;
    
    if_spiInit( file, SPIBR_6000KHZ );
	
	if (sd_Init( file ) < 0)
		return -2;
	
	if (sd_State( file ) < 0)
		return -3;
	
	sd_getDriveSize( file, &sc );
	file->sectorCount = sc / 512;
	if (sc % 512 != 0)
		file->sectorCount--;
	
	return 0;
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
esint8 if_readBuf
    (
    hwInterface *file,
    euint32 address,
    euint8 *buf
    )
    {
    if ((file->numDevice >= QNT_SPI_DEV) || (spi_devices[ file->numDevice ].spi_number >= QNT_SPI))
        return -1;
    
    if_spiInit( file, f_device_baud[ file->numDevice ] );
    
	return sd_readSector( file, address, buf, 512 );
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
esint8 if_setPos
    (
    hwInterface *file,
    euint32 address
    )
    {
	return 0;
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
void if_spiInit
    (
    hwInterface *iface,
    euint8 br_val
    )
    {
    if (   (f_cur_device_number != iface->numDevice)
        && (iface->numDevice < QNT_SPI_DEV)
        && (spi_devices[ iface->numDevice ].spi_number < QNT_SPI)
       )
        {
        f_device_baud[ iface->numDevice ] = br_val;
        f_if_spi_handles[ spi_devices[ iface->numDevice ].spi_number ].init(
            &spi_devices[ iface->numDevice ].spi_mode, br_val );
        }
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
euint8 if_spiSend
    (
    hwInterface *iface,
    euint8 outgoing
    )
    {
	euint8 incoming = 0;
	
	if ((iface->numDevice < QNT_SPI_DEV) && (spi_devices[ iface->numDevice ].spi_number < QNT_SPI))
	    {
	    if_spiInit( iface, f_device_baud[ iface->numDevice ] );
	    incoming = f_if_spi_handles[ spi_devices[ iface->numDevice ].spi_number ].send(
	        &spi_devices[ iface->numDevice ].spi_mode, outgoing );
	    }
	    
	return incoming;
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
void if_spiSendBuf
    (
    hwInterface *iface,
    euint16 bufsize,
    const euint8 *buf
    )
    {
	if ((iface->numDevice < QNT_SPI_DEV) && (spi_devices[ iface->numDevice ].spi_number < QNT_SPI))
	    {
	    if_spiInit( iface, f_device_baud[ iface->numDevice ] );
	    f_if_spi_handles[ spi_devices[ iface->numDevice ].spi_number ].sendbuf(
	        &spi_devices[ iface->numDevice ].spi_mode, bufsize, buf );
	    }
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
esint8 if_writeBuf
    (
    hwInterface *file,
    euint32 address,
    euint8 *buf
    )
    {
    if ((file->numDevice >= QNT_SPI_DEV) || (spi_devices[ file->numDevice ].spi_number >= QNT_SPI))
        return -1;
    
    if_spiInit( file, f_device_baud[ file->numDevice ] );
    
	return sd_writeSector( file, address, buf );
    }
/*------------------------------------------------------------------------------------------------*/
/*================================================================================================*/