/*****************************************************************************\
*              efs - General purpose Embedded Filesystem library              *
*          --------------------- -----------------------------------          *
*                                                                             *
* Filename : hcs12cw_spi.c                                                    *
* Description : This file contains the functions needed to use efs for        *
*               accessing files on an SD-card connected to an Freescale       *
*               HCS12 via SPI.                                                *
*                                                                             *
* This program is free software; you can redistribute it and/or               *
* modify it under the terms of the GNU General Public License                 *
* as published by the Free Software Foundation; version 2                     *
* of the License.                                                             *
                                                                              *
* This program is distributed in the hope that it will be useful,             *
* but WITHOUT ANY WARRANTY; without even the implied warranty of              *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               *
* GNU General Public License for more details.                                *
*                                                                             *
* As a special exception, if other files instantiate templates or             *
* use macros or inline functions from this file, or you compile this          *
* file and link it with other works to produce a work based on this file,     *
* this file does not by itself cause the resulting work to be covered         *
* by the GNU General Public License. However the source code for this         *
* file must still be made available in accordance with section (3) of         *
* the GNU General Public License.                                             *
*                                                                             *
* This exception does not invalidate any other reasons why a work based       *
* on this file might be covered by the GNU General Public License.            *
*                                                                             *
*                                                    (c)2006 Lennart Yseboodt *
*                                                    (c)2006 Michael De Nil   *
*                                                    (c)2007 Konstantin Les   *
\*****************************************************************************/


#include <hidef.h>      /* common defines and macros */
#include <mc9s12xdt256.h>     /* derivative information */
#pragma LINK_INFO DERIVATIVE "mc9s12xdt256"

#include "interfaces/hcs12cw_spi.h"
#include "interfaces/sd.h"


/*================================================================================================*/
#define QNT_SPI (2)

#define SPI_DEVNUM_UNDEFINED (255)

#define SS_MODE_NA (0)
#define SS_MODE_HW (1)
#define SS_MODE_SW (2)


#define NUMBER_ARRAY_ELEMENTS(a) (sizeof(a) / sizeof((a)[0]))

#define GET_SPI0(b) do {while (!SPI0SR_SPIF) ; (b) = SPI0DR;} while(0);
#define GET_SPI1(b) do {while (!SPI1SR_SPIF) ; (b) = SPI1DR;} while(0);
#define PUT_SPI0(b) do {while (!SPI0SR_SPTEF) ; SPI0DR = (b);} while(0);
#define PUT_SPI1(b) do {while (!SPI1SR_SPTEF) ; SPI1DR = (b);} while(0);
/*================================================================================================*/

/*================================================================================================*/
struct spi_config
    {
    unsigned char ss_mode;
    void (*assertCS)(void);
    void (*deassertCS)(void);
    unsigned char transfer_mode;
    };
struct spidev_config
    {
    unsigned char spi_number;
    struct spi_config spi_mode;
    };
/*================================================================================================*/

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
#include "if_cfg.c"
#define QNT_SPI_DEV (NUMBER_ARRAY_ELEMENTS( spi_devices ))
/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

/*================================================================================================*/
static void f_get_SPI0(euint16 len, euint16 bufsize, euint8 *buf);
static void f_get_SPI1(euint16 len, euint16 bufsize, euint8 *buf);
static void f_initialize_SPI0(const struct spi_config *spi_mode, unsigned char br_val);
static void f_initialize_SPI1(const struct spi_config *spi_mode, unsigned char br_val);
static void f_receivebuf_SPI0(const struct spi_config *spi_mode, euint16 size, euint16 buflen,
    euint8 *buf);
static void f_receivebuf_SPI1(const struct spi_config *spi_mode, euint16 size, euint16 buflen,
    euint8 *buf);
static euint8 f_send_SPI0(const struct spi_config *spi_mode, euint8 outgoing);
static euint8 f_send_SPI1(const struct spi_config *spi_mode, euint8 outgoing);
static void f_sendbuf_SPI0(const struct spi_config *spi_mode, euint16 bufsize, const euint8 *buf);
static void f_sendbuf_SPI1(const struct spi_config *spi_mode, euint16 bufsize, const euint8 *buf);
/*================================================================================================*/

/*================================================================================================*/
static unsigned char f_cur_device_number[ QNT_SPI ] = {SPI_DEVNUM_UNDEFINED, SPI_DEVNUM_UNDEFINED};
static unsigned char f_device_baud[ QNT_SPI_DEV ];
static const struct
    {
    void (*init)(const struct spi_config *, euint8);
    euint8 (*send)(const struct spi_config *, euint8);
    void (*sendbuf)(const struct spi_config *, euint16, const euint8 *);
    void (*receivebuf)(const struct spi_config *, euint16, euint16, euint8 *);
    } f_if_spi_handles[ QNT_SPI ] = 
    {
    {f_initialize_SPI0, f_send_SPI0, f_sendbuf_SPI0, f_receivebuf_SPI0},
    {f_initialize_SPI1, f_send_SPI1, f_sendbuf_SPI1, f_receivebuf_SPI1}
    };
/*================================================================================================*/

/*================================================================================================*/
/*------------------------------------------------------------------------------------------------*/
#if 0
static void f_get_SPI0
    (
    euint16 size,
    euint16 buflen,
    euint8 *buf
    )
    {
    euint8 c;
    __asm
        {
        tfr   D,X                                ; D=buf
        ldy   size                               ; Y=len
        
        brclr SPI0SR,#SPI0SR_SPTEF_MASK,*+0      ; while(!SPI0SR_SPTEF);
        ldab  #0xFF                              ; SPI0DR = 0xFF
        stab  SPI0DR
        
        bra   loop_ctrl
        loop:
        brclr SPI0SR,#SPI0SR_SPTEF_MASK,*+0      ; while(!SPI0SR_SPTEF);
        ldab  #0xFF                              ; SPI0DR = 0xFF
        stab  SPI0DR
        
        brclr SPI0SR,#SPI0SR_SPIF_MASK,*+0       ; while(!SPI0SR_SPIF);
        ldab  SPI0DR                             ; c=SPI0DR
        stab  c
        ldd   buflen
        beq   loop_ctrl                          ; if (buflen != 0) {
        decw  buflen                             ;   buflen--
        ldab  c
        stab  1,X+                               ;   *buf++=c;
                                                 ; }
        
        loop_ctrl:
        dey
        bne   loop
        
        brclr SPI0SR,#SPI0SR_SPIF_MASK,*+0       ; while(!SPI0SR_SPIF);
        ldab  SPI0DR                             ; *buf++=SPI0DR
        ldy   buflen
        beq   ret
        stab  0,X
        ret:
        }
    }
#endif
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
#if 0
static void f_get_SPI1
    (
    euint16 size,
    euint16 buflen,
    euint8 *buf
    )
    {
    __asm
        {
        tfr   D,X                                ;D=buf
        ldy   size                               ;Y=size
        
        brclr SPI1SR,#SPI1SR_SPTEF_MASK,*+0      ;while(!SPI1SR_SPTEF);
        ldab  #0xFF                              ;SPI1DR = 0xFF
        stab  SPI1DR
        
        bra   loop_ctrl
        loop:
        brclr SPI1SR,#SPI1SR_SPTEF_MASK,*+0      ;while(!SPI1SR_SPTEF);
        ldab  #0xFF                              ;SPI1DR = 0xFF
        stab  SPI1DR
        
        brclr SPI1SR,#SPI1SR_SPIF_MASK,*+0       ;while(!SPI1SR_SPIF);
        ldab  SPI1DR                             ;*buf++=SPI1DR
        stab  1,X+
        
        loop_ctrl:
        dey
        bne   loop
        
        brclr SPI1SR,#SPI1SR_SPIF_MASK,*+0       ;while(!SPI1SR_SPIF);
        ldab  SPI1DR                             ;*buf++=SPI1DR
        stab  0,X
        }
    }
#endif
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
static void f_initialize_SPI0
    (
    const struct spi_config *spi_mode, /* (in) - SPI mode */
    euint8 br_val                      /* (in) - value of SPI0BR register -> SPI baud rate */
    )
    {
    int spi_master_ok;
    
    SPI0CR1_CPHA = 0;
    SPI0CR1_SPE = 0;
    SPI0BR = br_val;
    SPI0CR2 = (spi_mode->ss_mode == SS_MODE_HW) ? SPI0CR2_MODFEN_MASK : 0;
    do
        {
        SPI0CR1_CPHA = 0;
        
        if (spi_mode->ss_mode == SS_MODE_HW)
            {
            unsigned char spi_status;
            SPI0CR1 = SPI0CR1_SPE_MASK | SPI0CR1_MSTR_MASK | SPI0CR1_SSOE_MASK |
                ((spi_mode->transfer_mode & 0x03) << 2);
            spi_status = SPI0SR;
            spi_master_ok = ((spi_status & SPI0SR_MODF_MASK) == 0);
            }
        else
            {
            SPI0CR1 = SPI0CR1_SPE_MASK | SPI0CR1_MSTR_MASK |
                ((spi_mode->transfer_mode & 0x03) << 2);
            spi_master_ok = 1;
            }
        }
    while (!spi_master_ok);
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
static void f_initialize_SPI1
    (
    const struct spi_config *spi_mode, /* (in) - SPI mode */
    euint8 br_val                      /* (in) - value of SPI0BR register -> SPI baud rate */
    )
    {
    int spi_master_ok;
    
    SPI1CR1_CPHA = 0;
    SPI1CR1_SPE = 0;
    SPI1BR = br_val;
    SPI1CR2 = (spi_mode->ss_mode == SS_MODE_HW) ? SPI1CR2_MODFEN_MASK : 0;
    do
        {
        SPI1CR1_CPHA = 0;
        
        if (spi_mode->ss_mode == SS_MODE_HW)
            {
            unsigned char spi_status;
            SPI1CR1 = SPI1CR1_SPE_MASK | SPI1CR1_MSTR_MASK | SPI1CR1_SSOE_MASK |
                ((spi_mode->transfer_mode & 0x03) << 2);
            spi_status = SPI1SR;
            spi_master_ok = ((spi_status & SPI1SR_MODF_MASK) == 0);
            }
        else
            {
            SPI1CR1 = SPI1CR1_SPE_MASK | SPI1CR1_MSTR_MASK |
                ((spi_mode->transfer_mode & 0x03) << 2);
            spi_master_ok = 1;
            }
        }
    while (!spi_master_ok);
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
static void f_receivebuf_SPI0
    (
    const struct spi_config *spi_mode,
    euint16 size,
    euint16 buflen,
    euint8 *buf
    )
    {
    unsigned char c;
    
    if (size == 0)
        return;
    
    
    /* SS pin functionality */
    if (spi_mode->ss_mode == SS_MODE_SW)
        spi_mode->assertCS();
    
#if 0
    f_get_SPI0( size, buflen, buf );
#endif

    PUT_SPI0( 0xFF );
    while (--size > 0)
        {
        PUT_SPI0( 0xFF );
        GET_SPI0( c );
        if (buflen > 0)
            {
            *buf++ = c;
            buflen--;
            }
        }
    GET_SPI0( c );
    if (buflen > 0)
        *buf = c;
        
    /* SS pin functionality */
    if (spi_mode->ss_mode == SS_MODE_SW)
        spi_mode->deassertCS();
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
static void f_receivebuf_SPI1
    (
    const struct spi_config *spi_mode,
    euint16 size,
    euint16 buflen,
    euint8 *buf
    )
    {
    unsigned char c;
    
    if (size == 0)
        return;
    
    
    /* SS pin functionality */
    if (spi_mode->ss_mode == SS_MODE_SW)
        spi_mode->assertCS();
    
#if 0
    f_get_SPI1( size, buflen, buf );
#endif

    PUT_SPI1( 0xFF );
    while (--size > 0)
        {
        PUT_SPI1( 0xFF );
        GET_SPI1( c );
        if (buflen > 0)
            {
            *buf++ = c;
            buflen--;
            }
        }
    GET_SPI1( c );
    if (buflen > 0)
        *buf = c;
        
    /* SS pin functionality */
    if (spi_mode->ss_mode == SS_MODE_SW)
        spi_mode->deassertCS();
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
static euint8 f_send_SPI0
    (
    const struct spi_config *spi_mode,
    euint8 outgoing
    )
    {
    euint8 incoming = 0;
    
    /* SS pin functionality */
    if (spi_mode->ss_mode == SS_MODE_SW)
        spi_mode->assertCS();
    
    PUT_SPI0( outgoing );
    GET_SPI0( incoming );
    
    /* SS pin functionality */
    if (spi_mode->ss_mode == SS_MODE_SW)
        spi_mode->deassertCS();
    
    return incoming;
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
static euint8 f_send_SPI1
    (
    const struct spi_config *spi_mode,
    euint8 outgoing
    )
    {
    euint8 incoming = 0;
    
    /* SS pin functionality */
    if (spi_mode->ss_mode == SS_MODE_SW)
        spi_mode->assertCS();
    
    PUT_SPI1( outgoing );
    GET_SPI1( incoming );
    
    /* SS pin functionality */
    if (spi_mode->ss_mode == SS_MODE_SW)
        spi_mode->deassertCS();
    
    return incoming;
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
static void f_sendbuf_SPI0
    (
    const struct spi_config *spi_mode,
    euint16 bufsize,
    const euint8 *buf
    )
    {
    t_if_cpusr cpu_sr;
    unsigned char rx_byte;
    
    if (bufsize == 0)
        return;
    
    
    /* SS pin functionality */
    if (spi_mode->ss_mode == SS_MODE_SW)
        spi_mode->assertCS();
    
    while (bufsize-- > 1)
        {
        PUT_SPI0( *buf );
        buf++;
        }
    IF_ENTER_CRITICAL();
    PUT_SPI0( *buf );
    while (!SPI0SR_SPTEF)
        {
        continue;
        }
    rx_byte = SPI0SR;
    rx_byte = SPI0DR;
    IF_EXIT_CRITICAL();
    
    GET_SPI0( rx_byte );

    /* SS pin functionality */
    if (spi_mode->ss_mode == SS_MODE_SW)
        spi_mode->deassertCS();
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
static void f_sendbuf_SPI1
    (
    const struct spi_config *spi_mode,
    euint16 bufsize,
    const euint8 *buf
    )
    {
    t_if_cpusr cpu_sr;
    unsigned char rx_byte;
    
    if (bufsize == 0)
        return;
    
    
    /* SS pin functionality */
    if (spi_mode->ss_mode == SS_MODE_SW)
        spi_mode->assertCS();
    
    while (bufsize-- > 1)
        {
        PUT_SPI1( *buf );
        buf++;
        }
    IF_ENTER_CRITICAL();
    PUT_SPI1( *buf );
    while (!SPI1SR_SPTEF)
        {
        continue;
        }
    rx_byte = SPI1SR;
    rx_byte = SPI1DR;
    IF_EXIT_CRITICAL();
    
    GET_SPI1( rx_byte );

    /* SS pin functionality */
    if (spi_mode->ss_mode == SS_MODE_SW)
        spi_mode->deassertCS();
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
void if_assert_device
    (
    hwInterface *iface
    )
    {
    spi_devices[ iface->numDevice ].spi_mode.assertCS();
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
void if_deassert_device
    (
    hwInterface *iface
    )
    {
    spi_devices[ iface->numDevice ].spi_mode.deassertCS();
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
esint8 if_flush
    (
    hwInterface *file
    )
    {
    esint8 ret_val = 0;
    
    if ((file->numDevice >= QNT_SPI_DEV) || (spi_devices[ file->numDevice ].spi_number >= QNT_SPI))
        return -1;
    
    if (if_take_device( file ) != 0)
        return -2;
    
    ret_val = sd_Flush( file );
    
    (void)if_release_device( file );
    
    return ret_val;
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
esint8 if_initInterface
    (
    hwInterface *file,
    eint8 *opts                        /* (in) - used SPI device number */
    )
    {
    esint8 ret_val = 0;
    euint32 sc;
    
    if (*opts >= QNT_SPI_DEV)
        return -1;
    
    file->numDevice = *opts;
    
    if (if_take_device( file ) != 0)
        return -2;
    
    if_spiInit( file, SPIBR_6000KHZ );
    
	if (sd_Init( file ) < 0)
	    {
	    ret_val = -3;
		goto ret_initInterface;
	    }
	
	if (sd_State( file ) < 0)
		{
	    ret_val = -4;
		goto ret_initInterface;
	    }
	
	(void)sd_getDriveSize( file, &sc );
	file->sectorCount = sc / 512;
	if (sc % 512 != 0)
		file->sectorCount--;
	
	/*if (!g_f)
	    goto ret_initInterface;
	aa = file->sectorCount * 512;
	for (ii = 0; (ii < aa); ii += 512)
	    {
	    euint8 resp;
	    
    	if_assert_device( file );
    	sd_Command( file, 29, (ii >> 16) & 0xFFFF, ii & 0xFFFF );
    	resp = sd_Resp8b( file );
    	if (resp != 0)
    	    {
    	    sd_Resp8bError( file, resp );
    	    return -4;
    	    }
    	    
    	while (if_spiSend( file, 0xFF ) == 0);
    	if_deassert_device( file );
	    }*/
	
	
ret_initInterface:
	if_release_device( file );
	
	return ret_val;
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
esint8 if_readBuf
    (
    hwInterface *file,
    euint32 address,
    euint8 *buf
    )
    {
    esint8 ret_val = -1;
    
    if ((file->numDevice >= QNT_SPI_DEV) || (spi_devices[ file->numDevice ].spi_number >= QNT_SPI))
        return -1;
    
    if (if_take_device( file ) != 0)
        return -2;
    
    if_spiInit( file, f_device_baud[ file->numDevice ] );
    
	ret_val = sd_readSector( file, address, buf, 512 );
	
	(void)if_release_device( file );
	
	return ret_val;
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
void if_spiReceiveBuf
    (
    hwInterface *iface,
    euint16 size,
    euint16 buflen,
    euint8 *buf
    )
    {
	if ((iface->numDevice < QNT_SPI_DEV) && (spi_devices[ iface->numDevice ].spi_number < QNT_SPI))
	    {
	    if_spiInit( iface, f_device_baud[ iface->numDevice ] );
	    f_if_spi_handles[ spi_devices[ iface->numDevice ].spi_number ].receivebuf(
	        &spi_devices[ iface->numDevice ].spi_mode, size, buflen, buf );
	    }
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
esint8 if_release_device
    (
    hwInterface *iface
    )
    {
    if (spi_devices[ iface->numDevice ].spi_number >= QNT_SPI)
        return -1;
    
    if_release_SPI( spi_devices[ iface->numDevice ].spi_number );
    
    return 0;
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
esint8 if_setPos
    (
    hwInterface *file,
    euint32 address
    )
    {
	return 0;
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
void if_spiInit
    (
    hwInterface *iface,
    euint8 br_val
    )
    {
    if (   (iface->numDevice < QNT_SPI_DEV)
        && (spi_devices[ iface->numDevice ].spi_number < QNT_SPI)
        && ((f_cur_device_number[ spi_devices[ iface->numDevice ].spi_number ] != iface->numDevice)
        || (f_device_baud[ iface->numDevice ] != br_val))
       )
        {
        f_device_baud[ iface->numDevice ] = br_val;
        f_if_spi_handles[ spi_devices[ iface->numDevice ].spi_number ].init(
            &spi_devices[ iface->numDevice ].spi_mode, br_val );
            
        f_cur_device_number[ spi_devices[ iface->numDevice ].spi_number ] = iface->numDevice;
        }
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
euint8 if_spiSend
    (
    hwInterface *iface,
    euint8 outgoing
    )
    {
	euint8 incoming = 0;
	
	if ((iface->numDevice < QNT_SPI_DEV) && (spi_devices[ iface->numDevice ].spi_number < QNT_SPI))
	    {
	    if_spiInit( iface, f_device_baud[ iface->numDevice ] );
	    incoming = f_if_spi_handles[ spi_devices[ iface->numDevice ].spi_number ].send(
	        &spi_devices[ iface->numDevice ].spi_mode, outgoing );
	    }
	    
	return incoming;
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
void if_spiSendBuf
    (
    hwInterface *iface,
    euint16 bufsize,
    const euint8 *buf
    )
    {
	if ((iface->numDevice < QNT_SPI_DEV) && (spi_devices[ iface->numDevice ].spi_number < QNT_SPI))
	    {
	    if_spiInit( iface, f_device_baud[ iface->numDevice ] );
	    f_if_spi_handles[ spi_devices[ iface->numDevice ].spi_number ].sendbuf(
	        &spi_devices[ iface->numDevice ].spi_mode, bufsize, buf );
	    }
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
esint8 if_take_device
    (
    hwInterface *iface
    )
    {
    if (spi_devices[ iface->numDevice ].spi_number >= QNT_SPI)
        return -1;
    
    if (if_take_SPI( spi_devices[ iface->numDevice ].spi_number ) != 0)
        return -2;
    
    return 0;
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
esint8 if_writeBuf
    (
    hwInterface *file,
    euint32 address,
    euint8 *buf
    )
    {
    esint8 ret_val;
    
    if ((file->numDevice >= QNT_SPI_DEV) || (spi_devices[ file->numDevice ].spi_number >= QNT_SPI))
        return -1;
    
    if (if_take_device( file ) != 0)
        return -2;
    
    if_spiInit( file, f_device_baud[ file->numDevice ] );
    
	ret_val = sd_writeSector( file, address, buf );
	
	(void)if_release_device( file );
	
	return ret_val;
    }
/*------------------------------------------------------------------------------------------------*/
/*================================================================================================*/