/*****************************************************************************\
*              efs - General purpose Embedded Filesystem library              *
*          --------------------- -----------------------------------          *
*                                                                             *
* Filename : sd.c                                                             *
* Revision : Initial developement                                             *
* Description : This file contains the functions needed to use efs for        *
*               accessing files on an SD-card.                                *
*                                                                             *
* This program is free software; you can redistribute it and/or               *
* modify it under the terms of the GNU General Public License                 *
* as published by the Free Software Foundation; version 2                     *
* of the License.                                                             *
                                                                              *
* This program is distributed in the hope that it will be useful,             *
* but WITHOUT ANY WARRANTY; without even the implied warranty of              *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               *
* GNU General Public License for more details.                                *
*                                                                             *
* As a special exception, if other files instantiate templates or             *
* use macros or inline functions from this file, or you compile this          *
* file and link it with other works to produce a work based on this file,     *
* this file does not by itself cause the resulting work to be covered         *
* by the GNU General Public License. However the source code for this         *
* file must still be made available in accordance with section (3) of         *
* the GNU General Public License.                                             *
*                                                                             *
* This exception does not invalidate any other reasons why a work based       *
* on this file might be covered by the GNU General Public License.            *
*                                                                             *
*                                                    (c)2006 Lennart Yseboodt *
*                                                    (c)2006 Michael De Nil   *
\*****************************************************************************/

#include "interfaces/sd.h"


/*================================================================================================*/
void f_wait_write_finish(hwInterface *iface, int finilize_write);
/*================================================================================================*/

/*================================================================================================*/
static const unsigned long f_invalid_address = 0xFFFFFFFFUL;
/*================================================================================================*/

/*================================================================================================*/
/*------------------------------------------------------------------------------------------------*/
void f_wait_write_finish
    (
    hwInterface *iface,
    int finilize_write
    )
    {
    if (iface->addressWritten != f_invalid_address)
        {
        while(if_spiSend( iface, 0xFF ) != 0xFF)
    	    {
            continue;
    	    }
    	(void)if_spiSend( iface, 0xFF );
    	
    	if (finilize_write)
    	    {
        	/* Stop Tran Token */
        	(void)if_spiSend( iface, 0xFD );
        	(void)if_spiSend( iface, 0xFF );
        	
        	while(if_spiSend( iface, 0xFF ) != 0xFF)
        	    {
                continue;
        	    }
        	(void)if_spiSend( iface, 0xFF );
        	
        	iface->addressWritten = f_invalid_address;
    	    }
        }
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
esint8 sd_Flush
    (
    hwInterface *iface
    )
    {
    if_assert_device( iface );
    f_wait_write_finish( iface, 1 );
    if_deassert_device( iface );
    return 0;
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
esint8 sd_Init
    (
    hwInterface *iface
    )
    {
	esint16 i;
	euint8 resp;
	
	iface->addressWritten = f_invalid_address;
	
	/* Try to send reset command up to 100 times */
	i = 100;
	if_assert_device( iface );
	do
	    {
		sd_Command( iface, 0, 0, 0 );
		resp = sd_Resp8b( iface );
		(void)if_spiSend( iface, 0xFF );
	    }
	while ((resp != 1) && i--);
	(void)if_spiSend( iface, 0xFF );
	if_deassert_device( iface );
	
	if (resp != 1)
	    {
		if(resp == 0xFF)
		    {
			return -1;
		    }
		else
		    {
			sd_Resp8bError( iface, resp );
			return -2;
		    }
	    }

	/* Wait till card is ready initialising (returns 0 on CMD1) */
	/* Try up to 32000 times. */
	i = 32000;
	if_assert_device( iface );
	do
	    {
		sd_Command( iface, 1, 0, 0 );
		resp = sd_Resp8b( iface );
		(void)if_spiSend( iface, 0xFF );
		
		if (resp != 0)
			sd_Resp8bError( iface, resp );
	    }
	while((resp == 1) && i--);
	(void)if_spiSend( iface, 0xFF );
	if_deassert_device( iface );
	
	if (resp != 0)
	    {
		sd_Resp8bError( iface, resp );
		return -3;
	    }
	
	/* Set SD-card sector length to 512 byte */
	if_assert_device( iface );
	sd_Command( iface, SDCMD_SET_BLOCKLEN, (512UL >> 16) & 0xFFFF, 512UL & 0xFFFF );
	resp = sd_Resp8b( iface );
	(void)if_spiSend( iface, 0xFF );
	if_deassert_device( iface );

	if (resp != 0)
	    {
	    sd_Resp8bError( iface, resp );
	    return -4;
	    }
	
	return 0;
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
esint8 sd_getDriveSize
    (
    hwInterface *iface,
    euint32 *drive_size
    )
    {
    /*
     * calculates size of card from CSD 
     * (extension by Martin Thomas, inspired by code from Holger Klabunde)
     */
    euint8 by;
    euint16 c_size;
	euint16 c_size_mult;
	euint8 cardresp;
	euint8 i;
	euint8 iob[ 16 ];
	euint16 read_bl_len;
	
	if_assert_device( iface );
	
	f_wait_write_finish( iface, 1 );
	
	sd_Command( iface, SDCMD_SEND_CSD, 0, 0 );
	
	do
	    {
		cardresp = sd_Resp8b( iface );
	    }
	while (cardresp != 0xFE);

	DBG( (TXT( "CSD:" )) );
	for (i = 0; (i < 16); i++)
	    {
		iob[ i ] = if_spiSend( iface, 0xFF );
		DBG( (TXT( " %02x" ), iob[ i ]) );
	    }
	DBG( (TXT( "\n" )) );

	(void)if_spiSend( iface, 0xFF );
	(void)if_spiSend( iface, 0xFF );
	
	(void)if_spiSend( iface, 0xFF );
	
	if_deassert_device( iface );
	
	c_size = iob[ 6 ] & 0x03; // bits 1..0
	c_size <<= 10;
	c_size += (euint16)iob[ 7 ] << 2;
	c_size += iob[ 8 ] >> 6;

	by = iob[ 5 ] & 0x0F;
	read_bl_len = 1;
	read_bl_len <<= by;

	by = iob[ 9 ] & 0x03;
	by <<= 1;
	by += iob[ 10 ] >> 7;
	
	c_size_mult = 1;
	c_size_mult <<= (2 + by);
	
	*drive_size = (euint32)(c_size + 1) * (euint32)c_size_mult * (euint32)read_bl_len;
	
	return 0;
}
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
esint8 sd_readSector
    (
    hwInterface *iface,
    euint32 address,
    euint8 *buf,
    euint16 len
    )
    {
    /*
     * WAIT ?? -- FIXME
     * CMDCMD
     * WAIT
     * CARD RESP
     * WAIT
     * DATA BLOCK IN
     * 		START BLOCK
     * 		DATA
     * 		CHKS (2B)
     */
     
	euint8 cardresp;
	euint16 fb_timeout = 0xFFFF;
	euint8 firstblock;
	euint32 place;
	esint8 ret_val = 0;

	/*DBG((TXT("sd_readSector::Trying to read sector %u and store it at %p.\n"),address,&buf[0]));*/
	place = 512 * address;
	
	if_assert_device( iface );
	
	f_wait_write_finish( iface, 1 );
	
	sd_Command( iface, SDCMD_READ_SINGLE_BLOCK, (euint16)(place >> 16), (euint16)place );
	
	cardresp = sd_Resp8b( iface ); /* Card response */ 

	/* Wait for startblock */
	do
	    {
		firstblock = sd_Resp8b( iface );
	    }
	while ((firstblock == 0xFF) && fb_timeout--);

	if ((cardresp != 0x00) || (firstblock != 0xFE))
	    {
		sd_Resp8bError( iface, firstblock );
		ret_val = -1;
		goto readSector_ret;
	    }
	
	if_spiReceiveBuf( iface, 512, len, buf );
	
	/* Checksum (2 byte) - ignore for now */
	(void)if_spiSend( iface, 0xFF );
	(void)if_spiSend( iface, 0xFF );
	
	(void)if_spiSend( iface, 0xFF );

readSector_ret:
    if_deassert_device( iface );
    
	return ret_val;
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
esint8 sd_State
    (
    hwInterface *iface
    )
    {
	eint16 value;
	
	if_assert_device( iface );
	
	f_wait_write_finish( iface, 1 );
	
	sd_Command( iface, 13, 0, 0 );
	value = sd_Resp16b( iface );
	
	(void)if_spiSend( iface, 0xFF );
	
	if_deassert_device( iface );
	

	switch(value)
	    {
		case 0x0000:
			return 1;
			break;
		case 0x0001:
			DBG( (TXT( "Card is Locked.\n" )) );
			break;
		case 0x0002:
			DBG( (TXT( "WP Erase Skip, Lock/Unlock Cmd Failed.\n" )) );
			break;
		case 0x0004:
			DBG( (TXT( "General / Unknown error -- card broken?.\n" )) );
			break;
		case 0x0008:
			DBG( (TXT( "Internal card controller error.\n" )) );
			break;
		case 0x0010:
			DBG( (TXT( "Card internal ECC was applied, but failed to correct the data.\n" )) );
			break;
		case 0x0020:
			DBG( (TXT( "Write protect violation.\n" )) );
			break;
		case 0x0040:
			DBG( (TXT( "An invalid selection, sectors for erase.\n" )) );
			break;
		case 0x0080:
			DBG( (TXT( "Out of Range, CSD_Overwrite.\n" )) );
			break;
		default:
			if (value > 0x00FF)
			    {
				sd_Resp8bError( iface, (euint8)(value >> 8) );
			    }
			else
			    {
				DBG( (TXT( "Unknown error: 0x%x (see SanDisk docs p5-14).\n" ), value) );
			    }
			break;
	    }
	return -1;
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
esint8 sd_writeSector
    (
    hwInterface *iface,
    euint32 address,
    euint8 *buf
    )
    {
    /* WAIT ?? -- FIXME
     * CMDWRITE
     * WAIT
     * CARD RESP
     * WAIT
     * DATA BLOCK OUT
     *      START BLOCK
     *      DATA
     *      CHKS (2B)
     * BUSY...
     */
    int is_nextSectorWrite;
	euint32 place;
	
	/*DBG((TXT("Trying to write %u to sector %u.\n"),(void *)&buf,address));*/
	place = 512 * address;
	
	if_assert_device( iface );
	
	is_nextSectorWrite = ((iface->addressWritten != f_invalid_address) &&
	    (iface->addressWritten + 1 == address));
    if (is_nextSectorWrite)
        {
    	f_wait_write_finish( iface, 0 );
        }
    else
        {
    	f_wait_write_finish( iface, 1 );
    	
    	sd_Command( iface, SDCMD_WRITE_MULTIPLE_BLOCK, (euint16)(place >> 16), (euint16)place );
    	(void)sd_Resp8b( iface ); /* Card response */
        }

    (void)if_spiSend( iface, 0xFC ); /* Start block */
	
	if_spiSendBuf( iface, 512, buf );
	
	(void)if_spiSend( iface, 0xFF ); /* Checksum part 1 */
	(void)if_spiSend( iface, 0xFF ); /* Checksum part 2 */

    iface->addressWritten = address;
    
	if_deassert_device( iface );
	
	/*DBG((TXT("Nopp'ed %u times.\n"),t));*/

	return 0;
    }
/*------------------------------------------------------------------------------------------------*/

void sd_Command(hwInterface *iface,euint8 cmd, euint16 paramx, euint16 paramy)
{
	//(void)if_spiSend(iface,0xff);
	
	(void)if_spiSend(iface,0x40 | cmd);
	(void)if_spiSend(iface,(euint8) (paramx >> 8)); /* MSB of parameter x */
	(void)if_spiSend(iface,(euint8) (paramx)); /* LSB of parameter x */
	(void)if_spiSend(iface,(euint8) (paramy >> 8)); /* MSB of parameter y */
	(void)if_spiSend(iface,(euint8) (paramy)); /* LSB of parameter y */

	(void)if_spiSend(iface,0x95); /* Checksum (should be only valid for first command (0) */

	(void)if_spiSend(iface,0xff); /* eat empty command - response */
}
/*****************************************************************************/

euint8 sd_Resp8b(hwInterface *iface)
{
	euint8 i;
	euint8 resp;
	
	/* Respone will come after 1 - 8 pings */
	for(i=0;i<8;i++){
		resp = if_spiSend(iface,0xff);
		if(resp != 0xff)
			return(resp);
	}
		
	return(resp);
}
/*****************************************************************************/

euint16 sd_Resp16b(hwInterface *iface)
{
	euint16 resp;
	
	resp = ( sd_Resp8b(iface) << 8 ) & 0xff00;
	resp |= if_spiSend(iface,0xff);
	
	return(resp);
}
/*****************************************************************************/

void sd_Resp8bError(hwInterface *iface,euint8 value)
{
	switch(value)
	{
		case 0x40:
			DBG((TXT("Argument out of bounds.\n")));
			break;
		case 0x20:
			DBG((TXT("Address out of bounds.\n")));
			break;
		case 0x10:
			DBG((TXT("Error during erase sequence.\n")));
			break;
		case 0x08:
			DBG((TXT("CRC failed.\n")));
			break;
		case 0x04:
			DBG((TXT("Illegal command.\n")));
			break;
		case 0x02:
			DBG((TXT("Erase reset (see SanDisk docs p5-13).\n")));
			break;
		case 0x01:
			DBG((TXT("Card is initialising.\n")));
			break;
		default:
			DBG((TXT("Unknown error 0x%x (see SanDisk docs p5-13).\n"),value));
			break;
	}
}
/*****************************************************************************/
/*================================================================================================*/
