/**
 * @file
 * Ethernet Interface Skeleton
 *
 */

/*
 * Copyright (c) 2001-2004 Swedish Institute of Computer Science.
 * All rights reserved. 
 * 
 * Redistribution and use in source and binary forms, with or without modification, 
 * are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission. 
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED 
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF 
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT 
 * SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
 * OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
 * IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY 
 * OF SUCH DAMAGE.
 *
 * This file is part of the lwIP TCP/IP stack.
 * 
 * Author: Adam Dunkels <adam@sics.se>
 *
 */

/*
 * This file is a skeleton for developing Ethernet network interface
 * drivers for lwIP. Add code to the low_level functions and do a
 * search-and-replace for the word "ethernetif" to replace it with
 * something that better describes your network interface.
 */

#include "lwip/opt.h"

#include "lwip/def.h"
#include "lwip/mem.h"
#include "lwip/pbuf.h"
#include "lwip/sys.h"
#include <lwip/stats.h>
#include <lwip/snmp.h>
#include "netif/etharp.h"

#include "lib/emac_bf/emac_bf.h"
#include "wdt.h"


/*================================================================================================*/
#define IFNAME0 'e'
#define IFNAME1 'n'
/*================================================================================================*/

/*================================================================================================*/
struct ethernetif
    {
    struct eth_addr *ethaddr;
    };
/*================================================================================================*/

/*================================================================================================*/
static void f_arp_timer(void *arg);
static err_t f_ethernetif_output(struct netif *netif, struct pbuf *p, struct ip_addr *ipaddr);
static void f_low_level_init(struct netif *netif);
static struct pbuf *f_low_level_input(struct netif *netif);
static err_t f_low_level_output(struct netif *netif, struct pbuf *p);
static portTASK_FUNCTION( f_th_ethernetif_input, param );
/*================================================================================================*/

/*================================================================================================*/
static portTickType f_guard_block_time = 250 / portTICK_RATE_MS;
static struct netif *f_netif = NULL;
/*================================================================================================*/

/*================================================================================================*/
/*------------------------------------------------------------------------------------------------*/
err_t ethernetif_init
    (
    struct netif *netif
    )
    {
    struct ethernetif *ethernetif;
    err_t ethifinit_err = ERR_OK;
    
    if ((ethernetif = mem_malloc( sizeof(struct ethernetif) )) == NULL)
        {
        LWIP_DEBUGF( NETIF_DEBUG, ("ethernetif_init: out of memory\n") );
        ethifinit_err = ERR_MEM;
        }
        
    if (ethifinit_err == ERR_OK)
        {
        netif->state = ethernetif;
        netif->name[ 0 ] = IFNAME0;
        netif->name[ 1 ] = IFNAME1;
        netif->output = f_ethernetif_output;
        netif->linkoutput = f_low_level_output;
        
        ethernetif->ethaddr = (struct eth_addr *)&(netif->hwaddr[ 0 ]);
        
        f_low_level_init( netif );
        etharp_init();
        sys_timeout( ARP_TMR_INTERVAL, f_arp_timer, NULL );
        }
    
    return ethifinit_err;
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
static void f_arp_timer
    (
    void *arg
    )
    {
    etharp_tmr();
    sys_timeout( ARP_TMR_INTERVAL, f_arp_timer, NULL );
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
static err_t f_ethernetif_output
    (
    struct netif *netif,
    struct pbuf *p,
    struct ip_addr *ipaddr
    )
    {
    return etharp_output( netif, p, ipaddr );
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
static void f_low_level_init
    (
    struct netif *netif
    )
    {
    const unsigned portBASE_TYPE ethif_task_prio = 3;//configMAX_PRIORITIES - 1;
    const unsigned portSHORT ethif_task_stacksize = 2048;
    const u16_t netif_mtu = 1500;
        
    netif->hwaddr_len = ETHARP_HWADDR_LEN;
    
    netif->hwaddr[ 0 ] = g_emac_hwaddr[ 0 ];
    netif->hwaddr[ 1 ] = g_emac_hwaddr[ 1 ];
    netif->hwaddr[ 2 ] = g_emac_hwaddr[ 2 ];
    netif->hwaddr[ 3 ] = g_emac_hwaddr[ 3 ];
    netif->hwaddr[ 4 ] = g_emac_hwaddr[ 4 ];
    netif->hwaddr[ 5 ] = g_emac_hwaddr[ 5 ];
    
    netif->mtu = netif_mtu;
    
    f_netif = netif;
    
    clear_wdt();
    while (emac_init() != e_EMACErr_NoErr)
        {
        clear_wdt();
        }
        
    xTaskCreate( f_th_ethernetif_input, "ETH_INT", ethif_task_stacksize, NULL, ethif_task_prio,
        NULL );
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
static struct pbuf *f_low_level_input
    (
    struct netif *netif
    )
    {
    static xSemaphoreHandle h_rx_smphr = NULL;
    struct pbuf *p = NULL;
    
    if (h_rx_smphr == NULL)
        vSemaphoreCreateBinary( h_rx_smphr );
        
    if (xSemaphoreTake( h_rx_smphr, f_guard_block_time ) == pdTRUE)
        {
        t_uint32 len;
        len = emac_input_length();
        if (len > 0)
            {
            struct pbuf *q;
                
#if ETH_PAD_SIZE
            len += ETH_PAD_SIZE;
#endif
            
            if ((p = pbuf_alloc( PBUF_RAW, len, PBUF_POOL )) != NULL)
                {
#if ETH_PAD_SIZE
                pbuf_header( p, -ETH_PAD_SIZE );
#endif

                emac_read_first( len, p->len, p->payload );
                for (q = p->next; (q != NULL); q = q->next)
                    {
                    emac_read_next( q->len, q->payload );
                    }
                    
#if ETH_PAD_SIZE
                pbuf_header( p, ETH_PAD_SIZE );
#endif
#if LINK_STATS
                lwip_stats.link.recv++;
#endif
                }
            else
                {
                vTaskDelay( 3 );
                }
            }
        else
            {
#if LINK_STATS
            lwip_stats.link.memerr++;
            lwip_stats.link.drop++;
#endif
            }
            
        xSemaphoreGive( h_rx_smphr );
        }
            
    
    return p;
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
static err_t f_low_level_output
    (
    struct netif *netif,
    struct pbuf *p
    )
    {
    static xSemaphoreHandle h_tx_smphr = NULL;
    err_t lloutput_err = ERR_OK;
    
    if (h_tx_smphr == NULL)
        vSemaphoreCreateBinary( h_tx_smphr );
    
#if ETH_PAD_SIZE
    pbuf_header( p, -ETH_PAD_SIZE );
#endif
    
        //sys_arch_protect();

    if (xSemaphoreTake( h_tx_smphr, f_guard_block_time ) == pdTRUE)
        {
        struct pbuf *q;
        
        for (q = p; (q != NULL); q = q->next)
            {
            if (emac_send( q->payload, q->len, (q->next == NULL) ) != e_EMACErr_NoErr)
                lloutput_err = ERR_IF;
            }
        xSemaphoreGive( h_tx_smphr );
        }
    else
        {
        lloutput_err = ERR_IF;
        }
        
        //sys_arch_unprotect( 0 );
        
        
#if ETH_PAD_SIZE
    pbuf_header( p, ETH_PAD_SIZE );
#endif

#if LINK_STATS
    lwip_stats.link.xmit++;
#endif
    
    return lloutput_err;
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
static portTASK_FUNCTION( f_th_ethernetif_input, param )
    {
    while (1)
        {
        struct eth_hdr *ethhdr;
        struct ethernetif *ethif;
        struct pbuf *p;
        
        static volatile int ii = 0;
        
        do
            {
            ethif = f_netif->state;
        
            if ((p = f_low_level_input( f_netif )) == NULL)
                emac_wait_input();
            }
        while (p == NULL);
        
#if LINK_STATS
        lwip_stats.link.recv++;
#endif

//        pbuf_free( p );
//        continue;
        
        ethhdr = p->payload;
        
        if ((ethhdr->dest.addr[ 2 ] != 0xFF) && (ethhdr->dest.addr[ 2 ] != 0xF7))
            ii++;
        
        switch (htons( ethhdr->type ))
            {
            case ETHTYPE_IP:
//                pbuf_free( p );
//                break;

//#if 0
                etharp_ip_input( f_netif, p );
                
                pbuf_header( p, -(s16_t)sizeof(struct eth_hdr) );
                
                if (ethhdr->dest.addr[ 0 ] != 0xFF)
                    ii++;
                
                f_netif->input( p, f_netif );
                break;
//#endif
                
            case ETHTYPE_ARP:
                etharp_arp_input( f_netif, ethif->ethaddr, p );
                break;
                
            default:
                pbuf_free( p );
                break;
            }
        }
    }
/*------------------------------------------------------------------------------------------------*/
